#include "agx_device_virtio.h"

#include <inttypes.h>
#include <sys/mman.h>

#include "drm-uapi/virtgpu_drm.h"

#define VIRGL_RENDERER_UNSTABLE_APIS 1
#include "vdrm.h"
#include "virglrenderer_hw.h"

struct agx_ccmd_ioctl_simple_req {
   struct vdrm_ccmd_req hdr;

   uint32_t cmd;
   uint8_t payload[];
};

struct agx_ccmd_ioctl_simple_rsp {
   struct vdrm_ccmd_rsp hdr;

   /* ioctl return value, interrupted syscalls are handled on the host without
    * returning to the guest.
    */
   int32_t ret;

   /* The output payload for IOC_RW ioctls, the payload is the same size as
    * msm_context_cmd_ioctl_simple_req.
    *
    * For IOC_W ioctls (userspace writes, kernel reads) this is zero length.
    */
   uint8_t payload[];
};

struct asahi_ccmd_get_params_req {
   struct vdrm_ccmd_req hdr;
   struct drm_asahi_get_params params;
};

struct asahi_ccmd_get_params_rsp {
   struct vdrm_ccmd_rsp hdr;
   int32_t ret;
   struct drm_asahi_params_global params;
};

struct asahi_ccmd_gem_new_req {
   struct vdrm_ccmd_req hdr;
   uint32_t flags;
   uint32_t bind_flags;
   uint32_t vm_id;
   uint32_t blob_id;
   uint64_t size;
   uint64_t addr;
};

struct asahi_ccmd_gem_bind_req {
   struct vdrm_ccmd_req hdr;
   uint32_t op;
   uint32_t flags;
   uint32_t vm_id;
   uint32_t res_id;
   uint64_t size;
   uint64_t addr;
};

struct asahi_ccmd_submit_req {
   struct vdrm_ccmd_req hdr;
   uint32_t cmd_type;
   uint32_t flags;
   uint64_t cmdbuf_size;
   uint64_t result_offset;
   uint64_t result_size;
   uint32_t barriers[2];
   uint32_t queue_id;
   uint32_t result_handle;
   uint32_t in_sync_count;
   uint32_t out_sync_count;

   uint8_t payload[];
};

struct asahi_ccmd_syncobj_wait_req {
   struct vdrm_ccmd_req hdr;
   uint32_t flags;
   uint32_t handle_count;
   int64_t timeout_nsecs;

   uint8_t payload[];
};

struct asahi_ccmd_syncobj_wait_rsp {
   struct vdrm_ccmd_rsp hdr;
   int32_t ret;
   uint32_t first_signaled;
};

struct asahi_ccmd_syncobj_create_req {
   struct vdrm_ccmd_req hdr;
   uint32_t flags;
};

struct asahi_ccmd_syncobj_create_rsp {
   struct vdrm_ccmd_rsp hdr;
   int32_t ret;
   uint32_t handle;
};

/*
 * Possible cmd types for "command stream", ie. payload of EXECBUF ioctl:
 */
enum agx_ccmd {
   AGX_CCMD_NOP = 1, /* No payload, can be used to sync with host */
   AGX_CCMD_IOCTL_SIMPLE,
   AGX_CCMD_GET_PARAMS,
   AGX_CCMD_GEM_NEW,
   AGX_CCMD_GEM_BIND,
   AGX_CCMD_SUBMIT,
   AGX_CCMD_SYNCOBJ_WAIT,
   AGX_CCMD_SYNCOBJ_CREATE,
   AGX_CCMD_LAST,
};

#ifdef __cplusplus
#define AGX_CCMD(_cmd, _len)                                                   \
   {                                                                           \
      .cmd = AGX_CCMD_##_cmd, .len = (_len),                                   \
   }
#else
#define AGX_CCMD(_cmd, _len)                                                   \
   (struct vdrm_ccmd_req)                                                      \
   {                                                                           \
      .cmd = AGX_CCMD_##_cmd, .len = (_len),                                   \
   }
#endif

/**
 * Helper for simple pass-thru ioctls
 */
int
agx_virtio_simple_ioctl(struct agx_device *dev, unsigned cmd, void *_req)
{
   struct vdrm_device *vdrm = dev->vdrm;
   unsigned req_len = sizeof(struct agx_ccmd_ioctl_simple_req);
   unsigned rsp_len = sizeof(struct agx_ccmd_ioctl_simple_rsp);

   req_len += _IOC_SIZE(cmd);
   if (cmd & IOC_OUT)
      rsp_len += _IOC_SIZE(cmd);

   uint8_t buf[req_len];
   struct agx_ccmd_ioctl_simple_req *req = (void *)buf;
   struct agx_ccmd_ioctl_simple_rsp *rsp;

   req->hdr = AGX_CCMD(IOCTL_SIMPLE, req_len);
   req->cmd = cmd;
   memcpy(req->payload, _req, _IOC_SIZE(cmd));

   rsp = vdrm_alloc_rsp(vdrm, &req->hdr, rsp_len);

   int ret = vdrm_send_req(vdrm, &req->hdr, true);
   if (ret) {
      fprintf(stderr, "simple_ioctl: vdrm_send_req failed\n");
      return ret;
   }

   if (cmd & IOC_OUT)
      memcpy(_req, rsp->payload, _IOC_SIZE(cmd));

   return rsp->ret;
}

static struct agx_bo *
agx_virtio_bo_alloc(struct agx_device *dev, size_t size,
                    size_t align, enum agx_bo_flags flags)
{
   struct agx_bo *bo;
   unsigned handle = 0;
   uint64_t ptr_gpu;

   size = ALIGN_POT(size, dev->params.vm_page_size);

   /* executable implies low va */
   assert(!(flags & AGX_BO_EXEC) || (flags & AGX_BO_LOW_VA));

   struct asahi_ccmd_gem_new_req req = {
      .hdr = AGX_CCMD(GEM_NEW, sizeof(req)),
      .size = size,
   };

   if (flags & AGX_BO_WRITEBACK)
      req.flags |= ASAHI_GEM_WRITEBACK;

   uint32_t blob_flags =
      VIRTGPU_BLOB_FLAG_USE_MAPPABLE | VIRTGPU_BLOB_FLAG_USE_SHAREABLE;

   req.bind_flags = ASAHI_BIND_READ;
   if (!(flags & AGX_BO_READONLY)) {
      req.bind_flags |= ASAHI_BIND_WRITE;
   }

   uint32_t blob_id = p_atomic_inc_return(&dev->next_blob_id);

   ASSERTED bool lo = (flags & AGX_BO_LOW_VA);

   struct util_vma_heap *heap;
   if (lo)
      heap = &dev->usc_heap;
   else
      heap = &dev->main_heap;

   simple_mtx_lock(&dev->vma_lock);
   ptr_gpu = util_vma_heap_alloc(heap, size + dev->guard_size,
                                 dev->params.vm_page_size);
   simple_mtx_unlock(&dev->vma_lock);
   if (!ptr_gpu) {
      fprintf(stderr, "Failed to allocate BO VMA\n");
      return NULL;
   }

   req.addr = ptr_gpu;
   req.blob_id = blob_id;
   req.vm_id = dev->vm_id;

   handle =
      vdrm_bo_create(dev->vdrm, size, blob_flags, blob_id, &req.hdr);
   if (!handle) {
      fprintf(stderr, "vdrm_bo_created failed\n");
      return NULL;
   }

   pthread_mutex_lock(&dev->bo_map_lock);
   bo = agx_lookup_bo(dev, handle);
   dev->max_handle = MAX2(dev->max_handle, handle);
   pthread_mutex_unlock(&dev->bo_map_lock);

   /* Fresh handle */
   assert(!memcmp(bo, &((struct agx_bo){}), sizeof(*bo)));

   bo->type = AGX_ALLOC_REGULAR;
   bo->size = size;
   bo->align = MAX2(dev->params.vm_page_size, align);
   bo->flags = flags;
   bo->dev = dev;
   bo->handle = handle;
   bo->prime_fd = -1;
   bo->blob_id = blob_id;
   bo->ptr.gpu = ptr_gpu;
   bo->vbo_res_id = vdrm_handle_to_res_id(dev->vdrm, handle);
   bo->guid = bo->handle; /* TODO: We don't care about guids */

   dev->ops.bo_mmap(bo);

   if (flags & AGX_BO_LOW_VA)
      bo->ptr.gpu -= dev->shader_base;

   assert(bo->ptr.gpu < (1ull << (lo ? 32 : 40)));

   return bo;
}

static int
agx_virtio_bo_bind(struct agx_device *dev, struct agx_bo *bo, uint64_t addr,
                   uint32_t flags)
{
   struct asahi_ccmd_gem_bind_req req = {
      .op = ASAHI_BIND_OP_BIND,
      .flags = flags,
      .vm_id = dev->vm_id,
      .res_id = bo->vbo_res_id,
      .size = bo->size,
      .addr = addr,
   };

   req.hdr.cmd = AGX_CCMD_GEM_BIND;
   req.hdr.len = sizeof(struct asahi_ccmd_gem_bind_req);

   int ret = vdrm_send_req(dev->vdrm, &req.hdr, false);
   if (ret) {
      fprintf(stderr, "DRM_IOCTL_ASAHI_GEM_BIND failed: %d (handle=%d)\n", ret,
              bo->handle);
   }

   return ret;
}

static void
agx_virtio_bo_mmap(struct agx_bo *bo)
{
   if (bo->ptr.cpu) {
      return;
   }
   /* If we have uploaded, we need to wait for host to handle that
    * before we can allow guest-side CPU access:
    */
   /* TODO (slp)
   if (virtio_bo->has_upload_seqno) {

      virtio_bo->has_upload_seqno = false;

      vdrm_flush(vdrm);
      vdrm_host_sync(vdrm, &(struct vdrm_ccmd_req) {
         .seqno = virtio_bo->upload_seqno,
      });
   }
   */

   bo->ptr.cpu = vdrm_bo_map(bo->dev->vdrm, bo->handle, bo->size);
   if (bo->ptr.cpu == MAP_FAILED) {
      bo->ptr.cpu = NULL;
      fprintf(stderr, "mmap failed: result=%p size=0x%llx fd=%i\n", bo->ptr.cpu,
              (long long)bo->size, bo->dev->fd);
   }
}

static ssize_t
agx_virtio_get_params(struct agx_device *dev, void *buf, size_t size)
{
   struct vdrm_device *vdrm = dev->vdrm;
   struct asahi_ccmd_get_params_req req;
   struct asahi_ccmd_get_params_rsp *rsp;

   memset(buf, 0, size);
   memset(&req.params, 0, sizeof(struct drm_asahi_get_params));

   req.params.size = size;
   req.hdr.cmd = AGX_CCMD_GET_PARAMS;
   req.hdr.len = sizeof(struct asahi_ccmd_get_params_req);

   rsp =
      vdrm_alloc_rsp(vdrm, &req.hdr, sizeof(struct asahi_ccmd_get_params_rsp));

   int ret = vdrm_send_req(vdrm, &req.hdr, true);
   if (ret)
      goto out;

   ret = rsp->ret;
   if (!ret) {
      memcpy(buf, &rsp->params, size);
      return size;
   }

out:
   return ret;
}

static int
agx_virtio_submit_single(struct agx_device *dev, uint32_t queue_id, bool feedback,
                         struct drm_asahi_cmd_compute *compute,
                         struct drm_asahi_cmd_render *render,
                         unsigned in_sync_count, struct drm_asahi_sync *in_syncs,
                         struct drm_asahi_sync *out_sync,
                         uint32_t result_handle, uint64_t batch_result_size,
                         uint64_t batch_result_off)
{
   if (compute) {
      // TODO (slp)
      fprintf(stderr, "Compute task not supported on asahi_virtio\n");
   }

   if (!render) {
      fprintf(stderr, "Missing render task\n");
      return ENODEV;
   }

   char *fragments = (char *)render->fragment_attachments;
   int fragment_count = render->fragment_attachment_count;

   size_t cmdbuf_size = sizeof(struct drm_asahi_cmd_render);
   size_t fragments_size =
      sizeof(struct drm_asahi_attachment) * fragment_count;

   int req_len = sizeof(struct asahi_ccmd_submit_req) +
                 cmdbuf_size + fragments_size;

   struct asahi_ccmd_submit_req *req =
      (struct asahi_ccmd_submit_req *)calloc(1, req_len);

   req->cmd_type = DRM_ASAHI_CMD_RENDER,
   req->flags = 0;
   req->cmdbuf_size = cmdbuf_size;
   req->queue_id = queue_id;
   req->result_handle = result_handle;
   req->result_offset = feedback ? (batch_result_off + batch_result_size) : 0;
   req->result_size = feedback ? batch_result_size : 0;

   struct drm_virtgpu_execbuffer_syncobj *vdrm_in_syncs =
      calloc(in_sync_count, sizeof(struct drm_virtgpu_execbuffer_syncobj));
   for (int i = 0; i < in_sync_count; i++) {
      vdrm_in_syncs[i].handle = in_syncs[i].handle;
   }

   struct drm_virtgpu_execbuffer_syncobj *vdrm_out_syncs =
      calloc(1, sizeof(struct drm_virtgpu_execbuffer_syncobj));
   vdrm_out_syncs[0].handle = out_sync[0].handle;

   char *ptr = (char *)&req->payload;
   memcpy(ptr, render, cmdbuf_size);
   ptr += cmdbuf_size;
   memcpy(ptr, (char *) fragments, fragments_size);

   req->hdr.cmd = AGX_CCMD_SUBMIT;
   req->hdr.len = req_len;

   unsigned rsp_len = sizeof(struct agx_ccmd_ioctl_simple_rsp);
   struct agx_ccmd_ioctl_simple_rsp *rsp =
      vdrm_alloc_rsp(dev->vdrm, &req->hdr, rsp_len);

   struct vdrm_execbuf_params p = {
      /* Signal the host we want to wait for the command to complete */
      .ring_idx = 1,
      .req = &req->hdr,
      .num_in_syncobjs = in_sync_count,
      .in_syncobjs = vdrm_in_syncs,
      .num_out_syncobjs = 1,
      .out_syncobjs = vdrm_out_syncs,
   };

   int ret = vdrm_execbuf(dev->vdrm, &p);
   if (ret) {
      ret = errno;
      goto out;
   }

   ret = rsp->ret;

out:
   if (ret == ENODEV)
      abort();

   free(vdrm_in_syncs);
   free(vdrm_out_syncs);
   free(req);

   return ret;
}

const agx_device_ops_t agx_virtio_device_ops = {
   .bo_alloc = agx_virtio_bo_alloc,
   .bo_bind = agx_virtio_bo_bind,
   .bo_mmap = agx_virtio_bo_mmap,
   .get_params = agx_virtio_get_params,
   .submit_single = agx_virtio_submit_single,
};

bool
agx_virtio_open_device(struct agx_device *dev)
{
   struct vdrm_device *vdrm;

   vdrm = vdrm_device_connect(dev->fd, 2);
   if (!vdrm) {
      fprintf(stderr, "could not connect vdrm\n");
      return false;
   }

   dev->vdrm = vdrm;
   dev->ops = agx_virtio_device_ops;
   return true;
}
